@@ -61,12 +61,14 @@ const struct value_string osmo_trau_frame_type_names[] = {
 	{ OSMO_TRAU16_FT_D145_SYNC,	"D145_SYNC" },
 	{ OSMO_TRAU16_FT_DATA_HR,	"DATA_HR" },
 	{ OSMO_TRAU16_FT_IDLE,		"IDLE" },
+	{ OSMO_TRAU16_FT_CONFIG,	"CONFIG" },
 	{ OSMO_TRAU8_SPEECH,		"8SPEECH" },
 	{ OSMO_TRAU8_DATA,		"8DATA" },
 	{ OSMO_TRAU8_OAM,		"8OAM" },
 	{ OSMO_TRAU8_AMR_LOW,		"8AMR_LOW" },
 	{ OSMO_TRAU8_AMR_6k7,		"8AMR_6k7" },
 	{ OSMO_TRAU8_AMR_7k4,		"8AMR_7k4" },
+	{ OSMO_TRAU8_CONFIG,		"8CONFIG" },
 	{ 0, NULL }
 };
 
@@ -238,30 +240,34 @@ static int encode8_handle_ta(ubit_t *trau_bits, const struct osmo_trau_frame *fr
 }
 
 /* TS 08.60 Section 3.1.1 */
-static int encode16_fr(ubit_t *trau_bits, const struct osmo_trau_frame *fr)
+static int encode16_fr(ubit_t *trau_bits, const struct osmo_trau_frame *fr, bool is_tfo)
 {
 	const ubit_t *cbits5;
 	int i;
 	int d_idx = 0;
 
-	switch (fr->type) {
-	case OSMO_TRAU16_FT_IDLE:
-		if (fr->dir == OSMO_TRAU_DIR_UL)
-			cbits5 = ft_idle_up_bits;
-		else
-			cbits5 = ft_idle_down_bits;
-		break;
-	case OSMO_TRAU16_FT_FR:
-		if (fr->dir == OSMO_TRAU_DIR_UL)
-			cbits5 = ft_fr_up_bits;
-		else
-			cbits5 = ft_fr_down_bits;
-		break;
-	case OSMO_TRAU16_FT_EFR:
-		cbits5 = ft_efr_bits;
-		break;
-	default:
-		return -EINVAL;
+	if (is_tfo) {
+		cbits5 = fr->c_bits;
+	} else {
+		switch (fr->type) {
+		case OSMO_TRAU16_FT_IDLE:
+			if (fr->dir == OSMO_TRAU_DIR_UL)
+				cbits5 = ft_idle_up_bits;
+			else
+				cbits5 = ft_idle_down_bits;
+			break;
+		case OSMO_TRAU16_FT_FR:
+			if (fr->dir == OSMO_TRAU_DIR_UL)
+				cbits5 = ft_fr_up_bits;
+			else
+				cbits5 = ft_fr_down_bits;
+			break;
+		case OSMO_TRAU16_FT_EFR:
+			cbits5 = ft_efr_bits;
+			break;
+		default:
+			return -EINVAL;
+		}
 	}
 
 	encode_sync16(trau_bits);
@@ -314,17 +320,21 @@ static int decode16_fr(struct osmo_trau_frame *fr, const ubit_t *trau_bits,
 }
 
 /* TS 08.60 Section 3.1.2 */
-static int encode16_amr(ubit_t *trau_bits, const struct osmo_trau_frame *fr)
+static int encode16_amr(ubit_t *trau_bits, const struct osmo_trau_frame *fr, bool is_tfo)
 {
-	const ubit_t *cbits5 = ft_amr_bits;
 	int i, d_idx;
 
 	encode_sync16(trau_bits);
 
-	/* C1 .. C5 */
-	memcpy(trau_bits + 17, cbits5 + 0, 5);
-	/* C6 .. C15 */
-	memcpy(trau_bits + 17 + 5, fr->c_bits + 5, 15 - 5);
+	if (is_tfo) {
+		/* C1 .. C15 */
+		memcpy(trau_bits + 17, fr->c_bits, 15);
+	} else {
+		/* C1 .. C5 */
+		memcpy(trau_bits + 17, ft_amr_bits, 5);
+		/* C6 .. C15 */
+		memcpy(trau_bits + 17 + 5, fr->c_bits + 5, 15 - 5);
+	}
 
 	trau_bits[32] = 1;
 	/* C16 .. C25 */
@@ -389,7 +399,7 @@ static int encode16_oam(ubit_t *trau_bits, const struct osmo_trau_frame *fr)
 	/* C1 .. C5 */
 	memcpy(trau_bits + 17, cbits5, 5);
 	/* C6 .. C15 */
-	memcpy(trau_bits + 17 + 5, fr->c_bits, 15 - 5);
+	memcpy(trau_bits + 17 + 5, fr->c_bits + 5, 15 - 5);
 
 	/* D1 .. D255 */
 	for (i = 32, d_idx = 0; i < 304; i += 16, d_idx += 15) {
@@ -514,52 +524,31 @@ static int decode16_hr(struct osmo_trau_frame *fr, const ubit_t *trau_bits,
 	return 0;
 }
 
-static struct osmo_trau_frame fr_idle_frame = {
-	.type = OSMO_TRAU16_FT_IDLE,
-	.dir = OSMO_TRAU_DIR_DL,
-	.t_bits = { 1, 1, 1, 1 },
-};
-#define TRAU_FRAME_BITS (40*8)
-static ubit_t encoded_idle_frame[TRAU_FRAME_BITS];
-static int dbits_initted = 0;
-
-/*! \brief return pointer to global buffer containing a TRAU idle frame */
-static ubit_t *trau_idle_frame(void)
-{
-	/* only initialize during the first call */
-	if (!dbits_initted) {
-		/* set all D-bits to 1 */
-		memset(&fr_idle_frame.d_bits, 0x01, 260);
-
-		memset(&fr_idle_frame.c_bits, 0x01, 25); /* spare are set to 1 */
-		/* set Downlink Idle Speech Frame pattern */
-		fr_idle_frame.c_bits[0] = 0; /* C1 */
-		fr_idle_frame.c_bits[1] = 1; /* C2 */
-		fr_idle_frame.c_bits[2] = 1; /* C3 */
-		fr_idle_frame.c_bits[3] = 1; /* C4 */
-		fr_idle_frame.c_bits[4] = 0; /* C5 */
-		/* set no Time Alignment pattern */
-		fr_idle_frame.c_bits[5] = 0; /* C6 */
-		fr_idle_frame.c_bits[6] = 0; /* C7 */
-		fr_idle_frame.c_bits[7] = 0; /* C8 */
-		fr_idle_frame.c_bits[8] = 0; /* C9 */
-		fr_idle_frame.c_bits[9] = 0; /* C10 */
-		fr_idle_frame.c_bits[10] = 0; /* C11 */
-		/* already set to 1, but maybe we need to modify it in the future */
-		fr_idle_frame.c_bits[11] = 1; /* C12 (UFE), good frame */
-		fr_idle_frame.c_bits[15] = 1; /* C16 (SP), no DTX */
-
-		encode16_fr(encoded_idle_frame, &fr_idle_frame);
-		dbits_initted = 1; /* set it to 1 to not call it again */
-	}
-	return encoded_idle_frame;
-}
-
 /* TS 08.60 Section 3.4 */
 static int encode16_idle(ubit_t *trau_bits, const struct osmo_trau_frame *fr)
 {
-	memcpy(trau_bits, trau_idle_frame(), 40*8);
-	return 40 * 8;
+	const ubit_t *cbits5;
+
+	if (fr->dir == OSMO_TRAU_DIR_UL)
+		cbits5 = ft_idle_up_bits;
+	else
+		cbits5 = ft_idle_down_bits;
+
+	encode_sync16(trau_bits);
+
+	/* C1 .. C5 */
+	memcpy(trau_bits + 17, cbits5 + 0, 5);
+	/* C6 .. C15 */
+	memcpy(trau_bits + 17 + 5, fr->c_bits + 5, 15 - 5);
+	/* all 1s in the place of D-bits */
+	memset(trau_bits + 32, 1, 310 - 32);
+	/* C16 .. C21 */
+	memcpy(trau_bits + 310, fr->c_bits + 15, 6);
+	/* T1 .. T4 */
+	memcpy(trau_bits + 316, fr->t_bits, 4);
+
+	/* handle timing adjustment */
+	return encode16_handle_ta(trau_bits, fr);
 }
 
 /* TS 08.60 Section 3.3.1 */
@@ -776,6 +765,57 @@ static int encode16_data_hr(ubit_t *trau_bits, const struct osmo_trau_frame *fr)
 	return 40 * 8;
 }
 
+/* TS 48.060 sections 5.1.5 and 5.5.1.5: generic configuration frames */
+
+static int decode16_gcf(struct osmo_trau_frame *fr, const ubit_t *trau_bits,
+			enum osmo_trau_frame_direction dir)
+{
+	int i, d_idx;
+
+	/* C1 .. C5 */
+	memcpy(fr->c_bits, trau_bits + 17, 5);
+
+	/* D1 .. D10 */
+	memcpy(fr->d_bits, trau_bits + 22, 10);
+	d_idx = 10;
+	/* D11 .. D265 */
+	for (i = 32; i < 304; i += 16) {
+		memcpy(fr->d_bits + d_idx, trau_bits + i + 1, 15);
+		d_idx += 15;
+	}
+	/* D266 .. D276 */
+	memcpy(fr->d_bits + d_idx, trau_bits + 305, 11);
+
+	/* T1 .. T4 */
+	memcpy(fr->t_bits, trau_bits + 316, 4);
+
+	return 0;
+}
+
+static int encode16_gcf(ubit_t *trau_bits, const struct osmo_trau_frame *fr)
+{
+	int i, d_idx;
+
+	encode_sync16(trau_bits);
+
+	/* C1 .. C5 */
+	memcpy(trau_bits + 17, fr->c_bits, 5);
+	/* D1 .. D10 */
+	memcpy(trau_bits + 22, fr->d_bits, 10);
+
+	/* D11 .. D276 */
+	for (i = 32, d_idx = 10; i <= 315; i += 16, d_idx += 15) {
+		trau_bits[i] = 1;
+		memcpy(trau_bits + i + 1, fr->d_bits + d_idx, 15);
+	}
+
+	/* T1 .. T4 */
+	memcpy(trau_bits + 316, fr->t_bits, 4);
+
+	/* handle timing adjustment */
+	return encode16_handle_ta(trau_bits, fr);
+}
+
 /* TS 08.61 Section 5.2.1.1 */
 static int decode8_hr(struct osmo_trau_frame *fr, const ubit_t *trau_bits,
 		      enum osmo_trau_frame_direction dir)
@@ -843,7 +883,7 @@ static ubit_t compute_odd_parity(const ubit_t *in, unsigned int num_bits)
 }
 
 /* TS 08.61 Section 5.2.1.1 */
-static int encode8_hr(ubit_t *trau_bits, const struct osmo_trau_frame *fr)
+static int encode8_hr(ubit_t *trau_bits, const struct osmo_trau_frame *fr, bool is_tfo)
 {
 	int i, d_idx = 0;
 
@@ -858,11 +898,15 @@ static int encode8_hr(ubit_t *trau_bits, const struct osmo_trau_frame *fr)
 	/* C1 .. C5 */
 	ubit_t *cbits_out = trau_bits + 1 * 8 + 1;
 	if (fr->dir == OSMO_TRAU_DIR_UL) {
-		cbits_out[0] = 0;
-		cbits_out[1] = 0;
-		cbits_out[2] = 0;
-		cbits_out[3] = 1;
-		cbits_out[4] = 0;
+		if (is_tfo) {
+			memcpy(cbits_out, fr->c_bits, 5);
+		} else {
+			cbits_out[0] = 0;
+			cbits_out[1] = 0;
+			cbits_out[2] = 0;
+			cbits_out[3] = 1;
+			cbits_out[4] = 0;
+		}
 	} else {
 		cbits_out[0] = 0;
 		cbits_out[1] = 0;
@@ -1001,15 +1045,15 @@ static int decode8_amr_67(struct osmo_trau_frame *fr, const ubit_t *trau_bits,
 	/* D8 .. D11 */
 	memcpy(fr->d_bits + d_idx, trau_bits + 2 * 8 + 4, 4);
 	d_idx += 4;
-	/* D12 .. D39 */
-	for (i = 3; i < 7; i++) {
+	/* D12 .. D32 */
+	for (i = 3; i < 6; i++) {
 		int offset = i * 8;
-		memcpy(fr->d_bits + d_idx, trau_bits + offset + 2, 7);
+		memcpy(fr->d_bits + d_idx, trau_bits + offset + 1, 7);
 		d_idx += 7;
 	}
 
-	/* D40 .. D137 */
-	for (i = 7; i < 20; i+= 2) {
+	/* D33 .. D137 */
+	for (i = 6; i < 20; i += 2) {
 		int offset = i * 8;
 		memcpy(fr->d_bits + d_idx, trau_bits + offset + 1, 15);
 		d_idx += 15;
@@ -1030,7 +1074,7 @@ static int encode8_amr_67(ubit_t *trau_bits, const struct osmo_trau_frame *fr)
 	trau_bits[3 * 8] = 1;
 	trau_bits[4 * 8] = 1;
 	trau_bits[5 * 8] = 0;
-	for (i = 5; i < 20; i += 2)
+	for (i = 6; i < 20; i += 2)
 		trau_bits[i * 8] = 1;
 
 	/* D1 .. D7 */
@@ -1041,15 +1085,15 @@ static int encode8_amr_67(ubit_t *trau_bits, const struct osmo_trau_frame *fr)
 	/* D8 .. D11 */
 	memcpy(trau_bits + 2 * 8 + 4, fr->d_bits + d_idx, 4);
 	d_idx += 4;
-	/* D12 .. D39 */
-	for (i = 3; i < 7; i++) {
+	/* D12 .. D32 */
+	for (i = 3; i < 6; i++) {
 		int offset = i * 8;
-		memcpy(trau_bits + offset + 2, fr->d_bits + d_idx, 7);
+		memcpy(trau_bits + offset + 1, fr->d_bits + d_idx, 7);
 		d_idx += 7;
 	}
 
-	/* D40 .. D137 */
-	for (i = 7; i < 20; i+= 2) {
+	/* D33 .. D137 */
+	for (i = 6; i < 20; i += 2) {
 		int offset = i * 8;
 		memcpy(trau_bits + offset + 1, fr->d_bits + d_idx, 15);
 		d_idx += 15;
@@ -1238,60 +1282,162 @@ static int encode8_oam(ubit_t *trau_bits, const struct osmo_trau_frame *fr)
 	return 20 * 8;
 }
 
+/* TS 48.061 section 5.2.1.3: generic configuration frames */
+
+static int decode8_gcf(struct osmo_trau_frame *fr, const ubit_t *trau_bits,
+		       enum osmo_trau_frame_direction dir)
+{
+	int i, d_idx = 0;
+
+	/* C1 .. C5 */
+	memcpy(fr->c_bits, trau_bits + 1 * 8 + 1, 5);
+	/* D1 .. D2 */
+	memcpy(fr->d_bits + d_idx, trau_bits + 1 * 8 + 6, 2);
+	d_idx += 2;
+	/* D3 .. D8 */
+	memcpy(fr->d_bits + d_idx, trau_bits + 2 * 8 + 2, 6);
+	d_idx += 6;
+	/* D9 .. D120 */
+	for (i = 3; i < 19; i++) {
+		memcpy(fr->d_bits + d_idx, trau_bits + i * 8 + 1, 7);
+		d_idx += 7;
+	}
+	/* D121 .. D125 */
+	memcpy(fr->d_bits + d_idx, trau_bits + 19 * 8 + 1, 5);
+
+	/* T1 .. T2 */
+	memcpy(fr->t_bits, trau_bits + 19 * 8 + 6, 2);
+
+	return 0;
+}
+
+static int encode8_gcf(ubit_t *trau_bits, const struct osmo_trau_frame *fr)
+{
+	int i, d_idx = 0;
+
+	/* sync pattern */
+	memset(trau_bits, 0, 8);
+	trau_bits[1 * 8 + 0] = 1;
+	trau_bits[2 * 8 + 0] = 0;
+	trau_bits[2 * 8 + 1] = 1;
+	for (i = 3; i < 20; i++)
+		trau_bits[i * 8] = 1;
+
+	/* C1 .. C5 */
+	memcpy(trau_bits + 1 * 8 + 1, fr->c_bits, 5);
+	/* D1 .. D2 */
+	memcpy(trau_bits + 1 * 8 + 6, fr->d_bits + d_idx, 2);
+	d_idx += 2;
+	/* D3 .. D8 */
+	memcpy(trau_bits + 2 * 8 + 2, fr->d_bits + d_idx, 6);
+	d_idx += 6;
+	/* D9 .. D120 */
+	for (i = 3; i < 19; i++) {
+		memcpy(trau_bits + i * 8 + 1, fr->d_bits + d_idx, 7);
+		d_idx += 7;
+	}
+	/* D121 .. D125 */
+	memcpy(trau_bits + 19 * 8 + 1, fr->d_bits + d_idx, 5);
+
+	/* T1 .. T2 */
+	memcpy(trau_bits + 19 * 8 + 6, fr->t_bits, 2);
+
+	/* handle timing adjustment */
+	return encode8_handle_ta(trau_bits, fr);
+}
+
 
 /*! Encode a TRAU frame from its decoded representation to a sequence of unpacked bits.
- *  \param[out] bits caller-allocated buffer for unpacked outpud bits
- *  \param[in] n_bits size of 'bits' oputput buffer in number of unpacked bits
+ *  \param[out] bits caller-allocated buffer for unpacked output bits
+ *  \param[in] n_bits size of 'bits' output buffer in number of unpacked bits
  *  \param[in] fr decoded representation of TRAU frame to be encoded
- *  \return 0 number of unpacked output bits generated; negative in case of error */
+ *  \returns number of unpacked output bits generated; negative in case of error
+ *
+ * This function exhibits a behavioral quirk which users need to be aware of:
+ * for many frame types, the first 5 bits out of user-provided fr->c_bits[]
+ * (corresponding to C1..C5) are ignored and replaced with internally supplied
+ * constant values, where the function "knows" what the correct frame type code
+ * should be.  This behavioral quirk is arguably a design defect, but it cannot
+ * be changed without breaking some existing applications that rely on this
+ * behavior and skip setting those bits themselves.  For TFO applications
+ * where C1..C5 may need to be modified (especially TFO-AMR where this C1..C5
+ * modification must be done before CRC computation), please use
+ * osmo_trau_frame_encode_tfo().
+ *
+ * The following list summarizes the behavior of the present function with
+ * regard to C1..C5 bits for different frame types:
+ *
+ * - For all TRAU-16k frame types in both UL and DL directions, with the
+ *   exception of newly added OSMO_TRAU16_FT_CONFIG, and for OSMO_TRAU8_SPEECH
+ *   (TRAU-8k for HRv1 speech) in UL direction, the first 5 bits of fr->c_bits[]
+ *   are ignored and replaced with internally supplied constant values.
+ *
+ * - For OSMO_TRAU8_SPEECH in DL direction, only fr->c_bits[3] is used to set
+ *   C4; constant values for C1..C3 and odd parity value for C5 are fixed
+ *   by the function.
+ *
+ * - For OSMO_TRAU8_DATA, OSMO_TRAU8_OAM and all AMR-8k frame types, as well as
+ *   newly added OSMO_TRAU16_FT_CONFIG and OSMO_TRAU8_CONFIG, user-supplied
+ *   fr->c_bits[] are always used.
+ *
+ * This unfortunate manipulation applies only to C1..C5 as listed above;
+ * for control bits C6 and higher (for frame types that have them),
+ * user-supplied fr->c_bits[] are always used.
+ */
 int osmo_trau_frame_encode(ubit_t *bits, size_t n_bits, const struct osmo_trau_frame *fr)
 {
+	size_t space_req;
+
 	/* check for sufficient space provided by caller in output buffer */
 	switch (fr->type) {
 	case OSMO_TRAU16_FT_FR:
 	case OSMO_TRAU16_FT_EFR:
 	case OSMO_TRAU16_FT_HR:
 	case OSMO_TRAU16_FT_AMR:
+	case OSMO_TRAU16_FT_IDLE:
+	case OSMO_TRAU16_FT_CONFIG:
 		/* timing alignment may happen: increased space requirement */
-		if (n_bits < 2 * 40 * 8 - 1)
-			return -ENOSPC;
+		if (fr->dir == OSMO_TRAU_DIR_DL && fr->dl_ta_usec > 0)
+			space_req = 2 * 40 * 8 - 1;
+		else
+			space_req = 1 * 40 * 8;
 		break;
 	case OSMO_TRAU16_FT_OAM:
-	case OSMO_TRAU16_FT_IDLE:
 	case OSMO_TRAU16_FT_DATA_HR:
 	case OSMO_TRAU16_FT_DATA:
 	case OSMO_TRAU16_FT_D145_SYNC:
 	case OSMO_TRAU16_FT_EDATA:
-		if (n_bits < 1 * 40 * 8)
-			return -ENOSPC;
+		space_req = 1 * 40 * 8;
 		break;
 	case OSMO_TRAU8_SPEECH:
 	case OSMO_TRAU8_AMR_LOW:
 	case OSMO_TRAU8_AMR_6k7:
 	case OSMO_TRAU8_AMR_7k4:
+	case OSMO_TRAU8_CONFIG:
 		/* timing alignment may happen: increased space requirement */
-		if (n_bits < 2 * 20 * 8 - 1)
-			return -ENOSPC;
+		if (fr->dir == OSMO_TRAU_DIR_DL && fr->dl_ta_usec > 0)
+			space_req = 2 * 20 * 8 - 1;
+		else
+			space_req = 1 * 20 * 8;
 		break;
 	case OSMO_TRAU8_DATA:
 	case OSMO_TRAU8_OAM:
-		if (n_bits < 1 * 20 * 8)
-			return -ENOSPC;
-		break;
-	case OSMO_TRAU_FT_NONE:
+		space_req = 1 * 20 * 8;
 		break;
 	default:
 		return -EINVAL;
 	}
+	if (n_bits < space_req)
+		return -ENOSPC;
 
 	switch (fr->type) {
 	case OSMO_TRAU16_FT_FR:
 	case OSMO_TRAU16_FT_EFR:
-		return encode16_fr(bits, fr);
+		return encode16_fr(bits, fr, false);
 	case OSMO_TRAU16_FT_HR:
 		return encode16_hr(bits, fr);
 	case OSMO_TRAU16_FT_AMR:
-		return encode16_amr(bits, fr);
+		return encode16_amr(bits, fr, false);
 	case OSMO_TRAU16_FT_OAM:
 		return encode16_oam(bits, fr);
 	case OSMO_TRAU16_FT_IDLE:
@@ -1304,8 +1450,10 @@ int osmo_trau_frame_encode(ubit_t *bits, size_t n_bits, const struct osmo_trau_f
 		return encode16_d145_sync(bits, fr);
 	case OSMO_TRAU16_FT_EDATA:
 		return encode16_edata(bits, fr);
+	case OSMO_TRAU16_FT_CONFIG:
+		return encode16_gcf(bits, fr);
 	case OSMO_TRAU8_SPEECH:
-		return encode8_hr(bits, fr);
+		return encode8_hr(bits, fr, false);
 	case OSMO_TRAU8_DATA:
 		return encode8_data(bits, fr);
 	case OSMO_TRAU8_OAM:
@@ -1316,7 +1464,71 @@ int osmo_trau_frame_encode(ubit_t *bits, size_t n_bits, const struct osmo_trau_f
 		return encode8_amr_67(bits, fr);
 	case OSMO_TRAU8_AMR_7k4:
 		return encode8_amr_74(bits, fr);
-	case OSMO_TRAU_FT_NONE:
+	case OSMO_TRAU8_CONFIG:
+		return encode8_gcf(bits, fr);
+	default:
+		return -EINVAL;
+	}
+}
+
+/*! Encode a TFO frame from its decoded representation to a sequence of unpacked bits.
+ *  \param[out] bits caller-allocated buffer for unpacked output bits
+ *  \param[in] n_bits size of 'bits' output buffer in number of unpacked bits
+ *  \param[in] fr decoded representation of TRAU frame to be encoded
+ *  \returns number of unpacked output bits generated; negative in case of error
+ *
+ * Compared to regular osmo_trau_frame_encode(), this TFO-specific TRAU frame
+ * encoding function restricts the set of possible frame types to those that
+ * are valid for TFO, restricts the direction to TRAU-UL, reduces the output
+ * space requirement (there is no timing alignment extension) and always uses
+ * all fr->c_bits provided by the application.
+ */
+int osmo_trau_frame_encode_tfo(ubit_t *bits, size_t n_bits, const struct osmo_trau_frame *fr)
+{
+	/* TFO frames are TRAU-UL only */
+	if (fr->dir != OSMO_TRAU_DIR_UL)
+		return -EINVAL;
+
+	/* check for sufficient space provided by caller in output buffer */
+	/* there is no timing alignment in TRAU-UL direction! */
+	switch (fr->type) {
+	case OSMO_TRAU16_FT_FR:
+	case OSMO_TRAU16_FT_EFR:
+	case OSMO_TRAU16_FT_AMR:
+	case OSMO_TRAU16_FT_CONFIG:
+		if (n_bits < 1 * 40 * 8)
+			return -ENOSPC;
+		break;
+	case OSMO_TRAU8_SPEECH:
+	case OSMO_TRAU8_AMR_LOW:
+	case OSMO_TRAU8_AMR_6k7:
+	case OSMO_TRAU8_AMR_7k4:
+	case OSMO_TRAU8_CONFIG:
+		if (n_bits < 1 * 20 * 8)
+			return -ENOSPC;
+		break;
+	default:
+		return -EINVAL;
+	}
+
+	switch (fr->type) {
+	case OSMO_TRAU16_FT_FR:
+	case OSMO_TRAU16_FT_EFR:
+		return encode16_fr(bits, fr, true);
+	case OSMO_TRAU16_FT_AMR:
+		return encode16_amr(bits, fr, true);
+	case OSMO_TRAU16_FT_CONFIG:
+		return encode16_gcf(bits, fr);
+	case OSMO_TRAU8_SPEECH:
+		return encode8_hr(bits, fr, true);
+	case OSMO_TRAU8_AMR_LOW:
+		return encode8_amr_low(bits, fr);
+	case OSMO_TRAU8_AMR_6k7:
+		return encode8_amr_67(bits, fr);
+	case OSMO_TRAU8_AMR_7k4:
+		return encode8_amr_74(bits, fr);
+	case OSMO_TRAU8_CONFIG:
+		return encode8_gcf(bits, fr);
 	default:
 		return -EINVAL;
 	}
@@ -1373,12 +1585,61 @@ int osmo_trau_frame_decode_16k(struct osmo_trau_frame *fr, const ubit_t *bits,
 	case TRAU_FT_D145_SYNC:
 		fr->type = OSMO_TRAU16_FT_D145_SYNC;
 		return decode16_data(fr, bits, dir);
+	case TRAU_FT_CONFIG_EXCH:
+		fr->type = OSMO_TRAU16_FT_CONFIG;
+		return decode16_gcf(fr, bits, dir);
 
 	default:
 		return -EINVAL;
 	}
 }
 
+/*! Decode/parse a 16k TFO frame from unpacked bits to decoded format.
+ *  \param[out] fr caller-allocated output data structure
+ *  \param[in] bits unpacked bits containing raw TFO frame; must be aligned
+ *  \returns 0 in case of success; negative on error
+ *
+ * TFO applications that need to decode TFO frames of TS 28.062 section 5.2.1
+ * (FR/EFR) or section 5.2.2.1 (AMR_TFO_16k) should use this function instead
+ * of regular osmo_trau_frame_decode_16k() in order to get appropriately
+ * modified handling of C1..C5 frame type bits:
+ *
+ * - C5 needs to be ignored because it is EMBED bit in TFO frames;
+ * - For interoperability with other potential TFO-AMR implementations
+ *   out in the world that aren't GSM AMR-FR, we need to recognize additional
+ *   C1..C4 bit patterns as being AMR equivalents: see Table 5.2.2.1-2 in
+ *   3GPP TS 28.062.
+ */
+int osmo_trau_frame_decode_tfo_16k(struct osmo_trau_frame *fr, const ubit_t *bits)
+{
+	uint8_t cbits4 = get_bits(bits, 17, 4);
+	/* We don't look at bit C5 because it is EMBED bit in TFO frames */
+
+	fr->type = OSMO_TRAU_FT_NONE;
+	fr->dir = OSMO_TRAU_DIR_UL;	/* TFO frames are modified TRAU-UL */
+	fr->dl_ta_usec = 0;
+
+	switch (cbits4) {
+	case TRAU_FT_FR_UP >> 1:
+		fr->type = OSMO_TRAU16_FT_FR;
+		return decode16_fr(fr, bits, fr->dir);
+	case TRAU_FT_EFR >> 1:
+		fr->type = OSMO_TRAU16_FT_EFR;
+		return decode16_fr(fr, bits, fr->dir);
+	case TRAU_FT_AMR >> 1:
+	case 0x4:	/* HR_AMR */
+	case 0x6:	/* UMTS_AMR_2 */
+	case 0xB:	/* OHR_AMR */
+		fr->type = OSMO_TRAU16_FT_AMR;
+		return decode16_amr(fr, bits, fr->dir);
+	case TRAU_FT_CONFIG_EXCH >> 1:
+		fr->type = OSMO_TRAU16_FT_CONFIG;
+		return decode16_gcf(fr, bits, fr->dir);
+	default:
+		return -EINVAL;
+	}
+}
+
 #define TRAU8_FT_AMR_NO_SPEECH_CMI	0x10	/* 1, 0, 0, 0, 0 */
 #define TRAU8_FT_AMR_NO_SPEECH_CMR	0x14	/* 1, 0, 1, 0, 0 */
 #define TRAU8_FT_AMR_475_515_590	0..7
@@ -1492,6 +1753,9 @@ int osmo_trau_frame_decode_8k(struct osmo_trau_frame *fr, const ubit_t *bits,
 			case 0x0B:
 				fr->type = OSMO_TRAU8_OAM;
 				return decode8_oam(fr, bits, dir);
+			case 0x1F:
+				fr->type = OSMO_TRAU8_CONFIG;
+				return decode8_gcf(fr, bits, dir);
 			}
 		} else {
 			/* Downlink */
@@ -1505,6 +1769,9 @@ int osmo_trau_frame_decode_8k(struct osmo_trau_frame *fr, const ubit_t *bits,
 			case 2:
 				fr->type = OSMO_TRAU8_OAM;
 				return decode8_oam(fr, bits, dir);
+			case 7:
+				fr->type = OSMO_TRAU8_CONFIG;
+				return decode8_gcf(fr, bits, dir);
 			}
 		}
 	} else if (is_amr_low(bits)) {
@@ -1521,4 +1788,87 @@ int osmo_trau_frame_decode_8k(struct osmo_trau_frame *fr, const ubit_t *bits,
 	return -EINVAL;
 }
 
+/*! Decode/parse an HRv1 TFO frame from unpacked bits to decoded format.
+ *  \param[out] fr caller-allocated output data structure
+ *  \param[in] bits 160 bits containing the lsb extracted from PCM samples;
+ *  must be aligned
+ *  \returns 0 in case of success; negative on error
+ *
+ * TFO applications that need to decode TFO frames of TS 28.062 section 5.3.1
+ * (GSM_HR codec) should use this function instead of regular
+ * osmo_trau_frame_decode_8k() in order to get these two behavior
+ * modifications:
+ *
+ * - The sync pattern check needs to be skipped because the sync pattern
+ *   may be partially overwritten by embedded TFO messages;
+ * - C5 needs to be ignored because it is EMBED bit in TFO frames.
+ *
+ * In the case of TFO, unlike regular TRAU-UL frames on Abis/Ater, there is
+ * no need to distinguish between HRv1 and AMR-8k frames by their respective
+ * sync patterns: per TS 28.062, these two TFO frame types never occur
+ * in the same bit position in TFO-superimposed PCM speech octets.
+ *
+ * Frame type bits C1..C4 are checked only to distinguish between regular
+ * HRv1 speech frames and Generic Configuration Frames of TS 28.062 Annex H:
+ * Rel5+ versions of TFO spec allow these GCFs to be used with all codecs
+ * if both ends of TFO connection implement Rel5+ version of TFO.
+ */
+int osmo_trau_frame_decode_tfo_hr1(struct osmo_trau_frame *fr, const ubit_t *bits)
+{
+	uint8_t cbits4 = get_bits(bits, 9, 4);
+	/* We don't look at bit C5 because it is EMBED bit in TFO frames */
+
+	fr->type = OSMO_TRAU_FT_NONE;
+	fr->dir = OSMO_TRAU_DIR_UL;	/* TFO frames are modified TRAU-UL */
+	fr->dl_ta_usec = 0;
+
+	switch (cbits4) {
+	case 0x01:
+		fr->type = OSMO_TRAU8_SPEECH;
+		return decode8_hr(fr, bits, fr->dir);
+	case 0x0F:
+		fr->type = OSMO_TRAU8_CONFIG;
+		return decode8_gcf(fr, bits, fr->dir);
+	default:
+		return -EINVAL;
+	}
+}
+
+/*! Decode/parse an AMR_TFO_8+8k frame from unpacked bits to decoded format.
+ *  \param[out] fr caller-allocated output data structure
+ *  \param[in] bits 160 bits containing the second lsb extracted from
+ *  PCM samples; must be aligned
+ *  \returns 0 in case of success; negative on error
+ *
+ * This function is provided for TFO applications that need to decode
+ * TFO frames of TS 28.062 section 5.2.2.2, format named AMR_TFO_8+8k.
+ * The second lsb extracted from PCM samples is the entity being decoded here.
+ * Compared to osmo_trau_frame_decode_8k(), the present function requires
+ * the direction to be TRAU-UL (as always the case in TFO) and accepts
+ * only AMR-8k frame types, or Generic Configuration Frames of TS 28.062
+ * Annex H that are used in Rel5+ version of TFO.
+ */
+int osmo_trau_frame_decode_tfo_amr_8k(struct osmo_trau_frame *fr, const ubit_t *bits)
+{
+	fr->type = OSMO_TRAU_FT_NONE;
+	fr->dir = OSMO_TRAU_DIR_UL;	/* TFO frames are modified TRAU-UL */
+	fr->dl_ta_usec = 0;
+
+	if (is_amr_low(bits)) {
+		fr->type = OSMO_TRAU8_AMR_LOW;
+		return decode8_amr_low(fr, bits, fr->dir);
+	} else if (is_amr_67(bits)) {
+		fr->type = OSMO_TRAU8_AMR_6k7;
+		return decode8_amr_67(fr, bits, fr->dir);
+	} else if (is_amr_74(bits)) {
+		fr->type = OSMO_TRAU8_AMR_7k4;
+		return decode8_amr_74(fr, bits, fr->dir);
+	} else if (is_hr(bits) && get_bits(bits, 9, 4) == 0x0F) {
+		fr->type = OSMO_TRAU8_CONFIG;
+		return decode8_gcf(fr, bits, fr->dir);
+	}
+
+	return -EINVAL;
+}
+
 /* }@ */
